use std::fmt::Display;
use std::ops::Sub;
use std::str::FromStr;

use getopts::Options;
use thiserror::Error;

const FLAG: &str = "FlÂG{E1nb38V9a3V5e86HzlY27UFAErIP5mA34kES9fbu}";

fn options() -> Options {
    let mut opts = Options::new();

    opts.reqopt(
        "f",
        "from",
        "calculate time starting from this time",
        "HH:MM",
    );
    opts.reqopt("t", "to", "calculate time until this time", "HH:MM");

    opts.optflag("h", "help", "print this help menu");

    opts
}

struct Timestamp {
    hour: i8,
    minute: i8,
}

#[derive(Error, Debug)]
enum TimestampParseError {
    #[error("Too many or too few separators ':'")]
    BadAmountOfSeparators,
    #[error("Time numbers t must fullfill 0 < t < 60 (was {0})")]
    InvalidTimeNumber(i8),
    #[error(transparent)]
    ParseInt(#[from] std::num::ParseIntError),
}

impl Sub for Timestamp {
    type Output = Self;

    fn sub(self, rhs: Self) -> Self::Output {
        if self.minute >= rhs.minute {
            Timestamp {
                hour: self.hour - rhs.hour,
                minute: self.minute - rhs.minute,
            }
        } else {
            Timestamp {
                hour: self.hour - rhs.hour - 1,
                minute: 60 + self.minute - rhs.minute,
            }
        }
    }
}

impl Display for Timestamp {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:02}:{:02}", self.hour, self.minute)
    }
}

impl FromStr for Timestamp {
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let parts: Vec<&str> = s.split(":").collect();
        if parts.len() != 2 {
            return Err(TimestampParseError::BadAmountOfSeparators);
        }

        let hour: i8 = parts[0].parse()?;
        let minute: i8 = parts[1].parse()?;

        if hour > 59 {
            return Err(TimestampParseError::InvalidTimeNumber(hour));
        }
        if minute > 59 {
            return Err(TimestampParseError::InvalidTimeNumber(minute));
        }

        let t = Timestamp { hour, minute };
        Ok(t)
    }

    type Err = TimestampParseError;
}

fn main() {
    let args: Vec<String> = std::env::args().collect();
    let program = args[0].clone();
    let opts = options();
    let matches = match opts.parse(&args[1..]) {
        Ok(m) => m,
        Err(f) => usage(&program, &opts),
    };
    let from_time: Timestamp = match matches.opt_str("from").expect("thats required").parse() {
        Ok(t) => t,
        Err(e) => {
            eprintln!("{e}");
            usage(&program, &opts)
        }
    };
    let to_time: Timestamp = match matches.opt_str("to").expect("thats required").parse() {
        Ok(t) => t,
        Err(e) => {
            eprintln!("{e}");
            usage(&program, &opts)
        }
    };

    println!("from\t\t: {from_time}");
    println!("to\t\t: {to_time}");
    println!("difference\t: {}", to_time - from_time)
}

fn usage(program: &str, opts: &Options) -> ! {
    let brief = format!("Usage: {} FILE [options]", program);
    print!("{}", opts.usage(&brief));
    std::process::exit(1)
}
